// EnvLayerSet.cpp : Implementation of CEnvLayerSet
//
// 11/19/03 Rich Williams - modified for use in SEEK
//

#include "EnvLayerSet.h"
#include "EnvLayer.h"
#include "EnvCell.h"
#include "Utilities.h"

#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <cstdlib>
#include <iostream>

#ifdef __GARPWIN
#include <direct.h>
#else
#include <unistd.h>
#endif

#include <scew.h>

// ================================================================================
// ================================================================================
// ================================================================================
EnvLayerSet::EnvLayerSet()
{
  strcpy(strId, "");
  strcpy(strTitle, "");
  strcpy(strFilename, "");
  strcpy(strDescription, "");

  strcpy(strMapUnits, "");
  strcpy(strCoordSys, "");

  dblXMin = dblXMax = dblYMin = dblYMax = dblCellSize = 0.0;

  intRows = intColumns = intCells = 0;

  intCount = 0;

  for (int i = 0; i < MAX_ENV_LAYERS; i++)
    objEnvLayers[i] = NULL;

  objSpeciesLayer = objMaskLayer = NULL;

  strcpy(strFilename, "");

  intPresencePointCount = 0;
  blnPresencePointCalculated = false;

  intLoadStatus = 0;

  lMaskedPoints = 0;
  lUnmaskedPoints = 0;
  lUniqueMaskedPoints = 0;
  lUniqueUnmaskedPoints = 0;
}

// ================================================================================
EnvLayerSet::~EnvLayerSet()
{
  for (int i = 0; i < MAX_ENV_LAYERS; i++)
    if (objEnvLayers[i])
      delete objEnvLayers[i];

  if (objMaskLayer)
    delete objMaskLayer;

  if (objSpeciesLayer)
    delete objSpeciesLayer;

  objSpeciesLayer = objMaskLayer = NULL;
}

// ================================================================================
int EnvLayerSet::size() const
{ return intCount; }

// ================================================================================
int EnvLayerSet::rows() const
{ return intRows; }

// ================================================================================
int EnvLayerSet::columns() const
{ return intColumns; }

// ================================================================================
char * EnvLayerSet::toXML()
{
  /* sample XML
    <EnvLayerSet Id="US_4km" Title="US_4km" Filename="US_4km.dat">
      Rows="623" Columns="1465"
      XMin="-126.020833" XMax="-64.979167"
      YMin="24.062500" YMax="50.020833" CellSize="0.041667"
      MapUnits="NULL" CoordSys="NULL">

      <EnvLayer Type="Species" Id="Layer1"></EnvLayer>
      <EnvLayer Type="Layer" Id="Layer2"></EnvLayer>
      <EnvLayer Type="Layer" Id="Layer3"></EnvLayer>
      <EnvLayer Type="Layer" Id="Layer4"></EnvLayer>
      <EnvLayer Type="Mask" Id="Layer5"></EnvLayer>
    </EnvLayerSet>
  */

  int i, size;
  char * strXML, strSpatialTag[2048], * strLayersTag, * strXMLLayer;

  size = 1024 * (intCount + 2);
  strLayersTag = new char[size];
  strXML       = new char[size + 8192];

  // create the Spatial tag
  sprintf(strSpatialTag, "Rows=\"%d\" Columns=\"%d\"\n XMin=\"%f\" XMax=\"%f\" YMin=\"%f\" YMax=\"%f\" CellSize=\"%f\"\n MapUnits=\"%s\" CoordSys=\"%s\"",
    intRows, intColumns, dblXMin, dblXMax, dblYMin, dblYMax, dblCellSize, strMapUnits, strCoordSys);

  // create layers tag
  strcpy(strLayersTag, "");

  // add species layer if exists
  if (objSpeciesLayer)
  {
    //strXMLLayer = objSpeciesLayer->toXML("Species");
    strXMLLayer = objSpeciesLayer->descriptionXML("Species");
    strcat(strLayersTag, strXMLLayer);
    strcat(strLayersTag, "\n");

    delete[] strXMLLayer;
  }

  // add regular layers
  for (i = 0; i < intCount; i++)
  {
    //strXMLLayer = objEnvLayers[i]->toXML("Layer");
    strXMLLayer = objEnvLayers[i]->descriptionXML("Layer");
    strcat(strLayersTag, strXMLLayer);
    strcat(strLayersTag, "\n");

    delete[] strXMLLayer;
  }

  // add mask layer if exists
  if (objMaskLayer)
  {
    //strXMLLayer = objMaskLayer->toXML("Mask");
    strXMLLayer = objMaskLayer->descriptionXML("Mask");
    strcat(strLayersTag, strXMLLayer);
    strcat(strLayersTag, "\n");

    delete[] strXMLLayer;
  }

  // put it all together (create EnvLayerSet tag)
  sprintf(strXML, "<EnvLayerSet Id=\"%s\" Title=\"%s\" Filename=\"%s\" %s>\n\n%s</EnvLayerSet>",
    strId, strTitle, strFilename, strSpatialTag, strLayersTag);

  if ((strlen(strSpatialTag) > 2048) ||
    (strlen(strLayersTag)  > size_t(size)) ||
    (strlen(strXML)        > size_t(size + 8192)))
    throw GarpException(82, "String size exceeded in EnvLayerSet::toXML()");

  delete[] strLayersTag;

  return strXML;
}

// ================================================================================
void EnvLayerSet::readXMLAttributes(scew_element const * pXMLSetNode)
{
  char * sName, * sValue;
  int i, n;

  scew_attribute * pXMLAttr;

  if (pXMLSetNode)
  {
    // get attribute list
    n = scew_attribute_count(pXMLSetNode);

    // process each attribute
    for (i = 0; i < n; i++)
    {
      pXMLAttr = scew_attribute_by_index(pXMLSetNode, i);

      sName = (char *) scew_attribute_name(pXMLAttr);
      sValue = (char *) scew_attribute_value(pXMLAttr);;

      if      (strcmp(sName, "Id") == 0)          strcpy(strId,          sValue);
      else if (strcmp(sName, "Title") == 0)       strcpy(strTitle,       sValue);
      else if (strcmp(sName, "Filename") == 0)    strcpy(strFilename,    sValue);
      else if (strcmp(sName, "Description") == 0) strcpy(strDescription, sValue);
      else if (strcmp(sName, "MapUnits") == 0)    strcpy(strMapUnits,    sValue);
      else if (strcmp(sName, "CoordSys") == 0)    strcpy(strCoordSys,    sValue);

      else if (strcmp(sName, "Rows") == 0)     intRows     = atoi(sValue);
      else if (strcmp(sName, "Columns") == 0)  intColumns  = atoi(sValue);
      else if (strcmp(sName, "XMin") == 0)     dblXMin     = atof(sValue);
      else if (strcmp(sName, "XMax") == 0)     dblXMax     = atof(sValue);
      else if (strcmp(sName, "YMin") == 0)     dblYMin     = atof(sValue);
      else if (strcmp(sName, "YMax") == 0)     dblYMax     = atof(sValue);
      else if (strcmp(sName, "CellSize") == 0) dblCellSize = atof(sValue);
    }

    intCells = intRows * intColumns;
  }

  else
    throw(1003, "Cannot read attributes of EnvLayerSet. XML node is empty.");

}

// ================================================================================

void EnvLayerSet::fromXMLFile(char * strInputFilename, int intLoadFlag)
{
  char S[256], * strXML;
  int lFileSize;
  int len = strlen(strInputFilename);
  intLoadStatus = intLoadFlag;

  lFileSize = GarpUtil::getFileSize(strInputFilename);
  if( lFileSize > 0 ) {
    char *strFile, *strPath, strCurDir[1024];
    strPath = new char[len+1];
    strFile = new char[len+1];
    GarpUtil::splitFileName(strInputFilename, strPath, strFile);
    if( *strPath != 0 ) {	// need to change current dir to dir of the xml file
#ifdef __GARPWIN					// because layers refer to files relative to xml file path
      _getcwd(strCurDir, 1024);
      _chdir(strPath);
#else
      getcwd(strCurDir, 1024);
      chdir(strPath);
#endif //__GARPWIN
    }
    FILE *inf = fopen(strFile, "r");
    if (inf)
    {
      strXML = new char[lFileSize + 32];
      strcpy(strXML, "");
      while (!feof(inf))
      {
        fscanf(inf, "%s\n", S);
        strcat(strXML, S);
        strcat(strXML, " ");		 // warning - the XML file can be no more than 32 lines long!
      }
      this->fromXML(strXML, intLoadFlag);
      fclose(inf);
      delete strXML;
    }
    if( *strPath != 0 ) {
#ifdef __GARPWIN
      _chdir(strCurDir);
#else
      chdir(strCurDir);
#endif //__GARPWIN
    }
    delete strPath;
    delete strFile;
  }
  else
    throw GarpException(83, "Could not open XML file");
}

// ================================================================================

void EnvLayerSet::toXMLFile(char * strOutputFilename)
{
  char *strXML = toXML();
  int iSize = strlen(strXML);
  FILE *outf = fopen(strOutputFilename, "w");
  if( outf ) {
    fwrite(strXML, sizeof(char), iSize, outf);
    fclose(outf);
    delete strXML;
  }
  else
  {
    delete strXML;
    throw GarpException(83, "Could not open XML file");
  }
}

// ================================================================================
void EnvLayerSet::fromXML(char * strXML, int intLoadFlag)
{
  int iResult, iStrLen;

  scew_tree* oTree = NULL;
  scew_parser* oParser = NULL;

  scew_element * pXMLNode, * pXMLRootNode;

  oParser = scew_parser_create();

    // Loads an XML file
  iStrLen = strlen(strXML);

  iResult = scew_parser_load_buffer(oParser, (const char *) strXML, iStrLen);

  if (iResult)
  {
    oTree = scew_parser_tree(oParser);

    // load data from XML Node
    pXMLRootNode = scew_tree_root(oTree);

    //scew_tree_save_file(oTree, "scewtree.out.xml");

    pXMLNode = scew_element_by_name_including_parent(pXMLRootNode, "EnvLayerSet");

    fromXML(pXMLNode, intLoadFlag);

    // Remember to free tree (scew_parser_free does not free it)
    // Frees the SCEW parser
    scew_parser_free(oParser);

    scew_tree_free(oTree); // TODO: find reason for memory bug

  }
  else
  {
    // Remember to free tree (scew_parser_free does not free it)
    // Frees the SCEW parser
    scew_tree_free(oTree);
    scew_parser_free(oParser);

    throw GarpException(22, "Could not load node from XML text.");
  }
}

// ================================================================================
void EnvLayerSet::fromXML(scew_element const * pXMLSetNode, int iLoadMode)
{
  long i, n;
  char * strType, * strAux;
  EnvLayer * objEnvLayer;

  unsigned int iNumItems;

  scew_element * pXMLNode, * pXMLLayerNode;
  scew_element ** pXMLLayerNodeList;

  //addref should have been called by the load method
  pXMLNode = scew_element_by_name_including_parent(pXMLSetNode, "EnvLayerSet");

  if (pXMLNode)
  {
    // parse EnvLayerSet node
    // read attributes

    readXMLAttributes(pXMLNode);

    // get list of EnvLayer nodes
    pXMLLayerNodeList = scew_element_list(pXMLNode, "EnvLayer", &iNumItems);

    // create Layers
    n = iNumItems;

    for (i = 0; i < n; i++)
    {

      pXMLLayerNode = pXMLLayerNodeList[i];

      // create new layer
      objEnvLayer = new EnvLayer;

      objEnvLayer->fromXML(pXMLLayerNode);

      // load the envlayer data if necessary
      // load mode 2 means load on memory (overwriting any on-disk setting)
      // load mode 3 means load on memory-mapped files (MMF)
      //    (overwriting any on-disk or on-memory settings)
      if (iLoadMode >= 2)
        objEnvLayer->loadData(iLoadMode);

      // to be used for dataset conversion
      //if (iLoadMode >= 1)
      //	objEnvLayer->loadText(objEnvLayer->strFilename, iLoadMode);

      strAux = GarpUtil::getXMLAttrAsString(pXMLLayerNode, "Type");
      strType = new char[strlen(strAux) + 1];
      strcpy(strType, strAux);

      // add layer following its type (default is layer)
      if (!strType || (strcmp(strType, "Layer") == 0))
        this->add(objEnvLayer);

      else if (strcmp(strType, "Species") == 0)
        this->addAsSpecies(objEnvLayer);

      else if (strcmp(strType, "Mask") == 0)
        this->addAsMask(objEnvLayer);

      delete strType;
    }

    // deallocate list
    scew_element_list_free(pXMLLayerNodeList);

  }
}

// ================================================================================
void EnvLayerSet::loadText(char * strLoadFilename, int iLoadMode)
{
  int i, intLayers;
  char S[256], S1[256];
  FILE * inf;
  EnvLayer * objEnvLayer;

  // DEPRECATED
  throw GarpException(1, "The method EnvLayerSet::loadText() is deprecated now. Use XML version instead.");

  strcpy(strFilename, strLoadFilename);
  inf = fopen(strFilename, "r");

  if (inf)
  {
    fscanf(inf, "%s %s", S, strId);
    fscanf(inf, "%s %s", S, strTitle);

    // ignore saved filename
    // use parameter instead
    fscanf(inf, "%s %s", S, S1);

    fscanf(inf, "%s %s", S, strMapUnits);
    fscanf(inf, "%s %s", S, strCoordSys);

    fscanf(inf, "%s", S); dblXMin = GarpUtil::readDoubleFromFile(inf);
    fscanf(inf, "%s", S); dblXMax = GarpUtil::readDoubleFromFile(inf);
    fscanf(inf, "%s", S); dblYMin = GarpUtil::readDoubleFromFile(inf);
    fscanf(inf, "%s", S); dblYMax = GarpUtil::readDoubleFromFile(inf);
    fscanf(inf, "%s", S); dblCellSize = GarpUtil::readDoubleFromFile(inf);

    fscanf(inf, "%s %d", S, &intRows);
    fscanf(inf, "%s %d", S, &intColumns);

    intCells = intRows * intColumns;

    fscanf(inf, "%s %d", S, &intLayers);

    for (i = 0; i < intLayers; i++)
    {
      objEnvLayer = new EnvLayer;

      fscanf(inf, "%s %s %s", S, objEnvLayer->strId, objEnvLayer->strFilename);

      if (iLoadMode)
      {
        objEnvLayer->loadText(objEnvLayer->strFilename, iLoadMode);

        // skip other two values
        GarpUtil::readDoubleFromFile(inf);
        GarpUtil::readDoubleFromFile(inf);
      }

      else
      {
        objEnvLayer->dblMinValue = GarpUtil::readDoubleFromFile(inf);
        objEnvLayer->dblMaxValue = GarpUtil::readDoubleFromFile(inf);
      }

      if (strcmp(S, "EnvLayer") == 0)
        add(objEnvLayer);

      else if (strcmp(S, "SpeciesLayer") == 0)
        addAsSpecies(objEnvLayer);

      else if (strcmp(S, "MaskLayer") == 0)
        addAsMask(objEnvLayer);

      else
      {
        char msg[256];
        sprintf(msg, "Environmental layer type %s not recognized", S);
        throw GarpException(1, msg);
      }

    }

    // close file
    fclose(inf);
  }
  else
  {
    char msg[256];
    sprintf(msg, "Environmental layer set metadata <%s> could not be opened for reading", strFilename);
    throw GarpException(1, msg);
  }
}

// ================================================================================
void EnvLayerSet::saveText(char * strSaveFilename)
{
  FILE * outf;
  int i, intLayers;

  if ((strcmp(strSaveFilename, "NULL") == 0) ||
    (strcmp(strSaveFilename, "") == 0))
    throw GarpException(2, "Cannot save dataset with an empty filename");

  strcpy(strFilename, strSaveFilename);
  outf = fopen(strFilename, "w");

  if (outf)
  {
    fprintf(outf, "%s %s\n", "Id", strId);
    fprintf(outf, "%s %s\n", "Title", strTitle);
    fprintf(outf, "%s %s\n", "Filename", strFilename);

    fprintf(outf, "%s %s\n", "MapUnits", strMapUnits);
    fprintf(outf, "%s %s\n", "CoordSys", strCoordSys);

    fprintf(outf, "%s %f\n", "MinX", dblXMin);
    fprintf(outf, "%s %f\n", "MaxX", dblXMax);
    fprintf(outf, "%s %f\n", "MinY", dblYMin);
    fprintf(outf, "%s %f\n", "MaxY", dblYMax);
    fprintf(outf, "%s %f\n", "CellSize", dblCellSize);

    fprintf(outf, "%s %d\n", "Rows", intRows);
    fprintf(outf, "%s %d\n", "Columns", intColumns);

    // the layers attribute should take into account
    // both mask and species layers if available
    intLayers = intCount + ((objSpeciesLayer)? 1:0) + ((objMaskLayer)? 1:0);
    fprintf(outf, "%s %d\n", "Layers", intLayers);

    if (objSpeciesLayer)
      fprintf(outf, "SpeciesLayer %s\n", objSpeciesLayer->getDescriptionText());

    for (i = 0; i < intCount; i++)
      fprintf(outf, "EnvLayer %s\n", objEnvLayers[i]->getDescriptionText());

    if (objMaskLayer)
      fprintf(outf, "MaskLayer %s\n", objMaskLayer->getDescriptionText());

    fclose(outf);
  }
  else
  {
    char msg[256];
    sprintf(msg, "Environmental layer metadata <%s> could not be opened for writing", strFilename);
    throw GarpException(1, msg);
  }
}

// ================================================================================
bool EnvLayerSet::isCompatible(const EnvLayer& objAnotherEnvLayer)
{
  bool result = true;

  if ((GarpUtil::equalEps(objAnotherEnvLayer.dblXMin, 0.0, EPS)) &&
    (GarpUtil::equalEps(objAnotherEnvLayer.dblXMax, 0.0, EPS)) &&
    (GarpUtil::equalEps(objAnotherEnvLayer.dblYMin, 0.0, EPS)) &&
    (GarpUtil::equalEps(objAnotherEnvLayer.dblYMax, 0.0, EPS)) &&
    (GarpUtil::equalEps(objAnotherEnvLayer.dblCellSize, 0.0, EPS)) &&
    (GarpUtil::equalEps(objAnotherEnvLayer.intRows, 0.0, EPS)) &&
    (GarpUtil::equalEps(objAnotherEnvLayer.intColumns, 0.0, EPS)))

    // env layer was initialized using short description
    // so the env layer properties were not initialized
    // they contain just zeros
    // therefore, it is considered compatible with the EnvLayerSet
    return true;

  else
  {
    result &= GarpUtil::equalEps(dblXMin, objAnotherEnvLayer.dblXMin, EPS);
    result &= GarpUtil::equalEps(dblXMax, objAnotherEnvLayer.dblXMax, EPS);
    result &= GarpUtil::equalEps(dblYMin, objAnotherEnvLayer.dblYMin, EPS);
    result &= GarpUtil::equalEps(dblYMax, objAnotherEnvLayer.dblYMax, EPS);
    result &= GarpUtil::equalEps(dblCellSize, objAnotherEnvLayer.dblCellSize, EPS);
    result &= GarpUtil::equalEps(intRows, objAnotherEnvLayer.intRows, EPS);
    result &= GarpUtil::equalEps(intColumns, objAnotherEnvLayer.intColumns, EPS);
  }

  return result;

  return ( (dblXMin == objAnotherEnvLayer.dblXMin) &&
       (dblXMax == objAnotherEnvLayer.dblXMax) &&
       (dblYMin == objAnotherEnvLayer.dblYMin) &&
       (dblYMax == objAnotherEnvLayer.dblYMax) &&
       (dblCellSize == objAnotherEnvLayer.dblCellSize) &&
       (intRows == objAnotherEnvLayer.intRows) &&
       (intColumns == objAnotherEnvLayer.intColumns) );
}

// ================================================================================
void EnvLayerSet::metadataFromEnvLayer(const EnvLayer& objEnvLayer)
{
  /*
  if ((!GarpUtil::equalEps(objEnvLayer.dblXMin, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.dblXMax, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.dblYMin, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.dblYMax, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.dblCellSize, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.intRows, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.intColumns, 0.0, EPS)))
  */

  if ((!GarpUtil::equalEps(objEnvLayer.dblXMin, objEnvLayer.dblXMax, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.dblYMin, objEnvLayer.dblYMax, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.dblCellSize, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.intRows, 0.0, EPS)) &&
    (!GarpUtil::equalEps(objEnvLayer.intColumns, 0.0, EPS)))
  {
    dblXMin		= objEnvLayer.dblXMin;
    dblXMax		= objEnvLayer.dblXMax;
    dblYMin		= objEnvLayer.dblYMin;
    dblYMax		= objEnvLayer.dblYMax;
    dblCellSize = objEnvLayer.dblCellSize;
    intRows		= objEnvLayer.intRows;
    intColumns	= objEnvLayer.intColumns;
    intCells	= objEnvLayer.intCells;

    if (strcmp(strMapUnits, "") == 0)
      strcpy(strMapUnits, objEnvLayer.strMapUnits);

    if (strcmp(strCoordSys, "") == 0)
      strcpy(strCoordSys, objEnvLayer.strCoordSys);
  }
}

// ================================================================================
void EnvLayerSet::resetWithMaskBytes(EnvLayer * objEnvLayer, BYTE bytValue)
{
  int i;
  BYTE bytMaskValue;

  for (i = 0; i < intCells; i++)
  {
    bytMaskValue = objMaskLayer->get(i);
    if (bytMaskValue == MASK_VALUE)
      objEnvLayer->set(i, MASK_VALUE);
    else
      objEnvLayer->set(i, bytValue);
  }
}

// ================================================================================
void EnvLayerSet::metadataToEnvLayer(EnvLayer * objEnvLayer)
{
  objEnvLayer->dblXMin     = dblXMin;
  objEnvLayer->dblXMax     = dblXMax;
  objEnvLayer->dblYMin     = dblYMin;
  objEnvLayer->dblYMax     = dblYMax;
  objEnvLayer->dblCellSize = dblCellSize;
  objEnvLayer->intRows     = intRows;
  objEnvLayer->intColumns  = intColumns;
  objEnvLayer->intCells    = intCells;

  strcpy(objEnvLayer->strMapUnits, strMapUnits);
  strcpy(objEnvLayer->strCoordSys, strCoordSys);
}

// ================================================================================
void EnvLayerSet::swapLayers(int iFirstId, int iSecondId)
{
  EnvLayer * pAux;

  if (objEnvLayers[iFirstId] && objEnvLayers[iSecondId])
  {
    pAux = objEnvLayers[iFirstId];
    objEnvLayers[iFirstId] = objEnvLayers[iSecondId];
    objEnvLayers[iSecondId] = pAux;
  }
  else
    throw GarpException(1, "One of the layers selected for swap are empty.");
}

// ================================================================================
void EnvLayerSet::removeLayer(int iIndex)
{
  if (objEnvLayers[iIndex])
  {
    delete objEnvLayers[iIndex];
    for (int i = iIndex; i < intCount - 1; i++)
      objEnvLayers[i] = objEnvLayers[i + 1];
    objEnvLayers[--intCount] = NULL;
  }
  else
    throw GarpException(1, "Could not remove layer. It does not exist.");
}

// ================================================================================
void EnvLayerSet::add(EnvLayer * layer)
{
  if (layer)
  {
    if (intCount < MAX_ENV_LAYERS)
    {
      if (intCount)
      {
        // there is at least one layer in this set: must check compatibility
        if (isCompatible(*layer))
        {
          // add layer
          objEnvLayers[intCount++] = layer;
        }
        else
          throw GarpException(1, "EnvLayer object being added is not compatible to the EnvLayerSet object");
      }
      else
      {
        // there is no layers in this set: must take metadata from first layer
        metadataFromEnvLayer(*layer);

        // add this layer as the first layer
        objEnvLayers[intCount++] = layer;
      }
    }

    else
      throw GarpException(1, "Max number of layers per set reached");
  }
  else
    throw GarpException(1, "Cannot add null pointer as EnvLayer");
}

// ================================================================================
void EnvLayerSet::addAsSpecies(EnvLayer * layer)
{
  // layer can be null
  objSpeciesLayer = layer;
}

// ================================================================================
void EnvLayerSet::addAsMask(EnvLayer * layer)
{
  if (objMaskLayer)
    delete objMaskLayer;

  // layer can be null
  objMaskLayer = layer;
}

// ================================================================================
bool EnvLayerSet::masked(int intPos)
{
  BYTE val;

  if (objMaskLayer)
  {
    val = objMaskLayer->get(intPos);
    return ((int) val == MASK_VALUE);
  }
  else
    return false;
    // throw GarpException(1, "No mask EnvLayer defined");
}

// ================================================================================
EnvCell * EnvLayerSet::getCell(int intPos)
{
  bool blnStatus;
  BYTE * values;
  EnvCell * cell;
  double x, y;

  objEnvLayers[0]->getCoordinates(intPos, x, y);

  values = new BYTE[intCount + 2];
  cell = new EnvCell(intCount + 2, values, intPos, x, y);
  blnStatus = getValue(intPos, cell);

  // check if cell is masked
  if (blnStatus)
    // it is not masked
    return cell;
  else
  {
    // yes: it is masked: return NULL
    delete cell;
    delete[] values;
    return NULL;
  }
}

// ================================================================================
bool EnvLayerSet::getValue(int intPos, EnvCell * cell)
{
  int i;
  bool blnMasked;
  BYTE bytMaskValue;


  if (objMaskLayer)
  {
    // get mask layer value
    bytMaskValue = objMaskLayer->get(intPos);
    blnMasked = bytMaskValue == MASK_VALUE;
  }
  else
  {
    // there is no mask
    bytMaskValue = 0;
    blnMasked = false;
  }


  // check if point is masked
  if (!blnMasked)
  {
    // put species value as the first value
    if (objSpeciesLayer)
      cell->values[0] = objSpeciesLayer->get(intPos);
    else
      cell->values[0] = 255;

    // point is not masked: get other layer values
    for (i = 0; i < intCount; i++)
      cell->values[i + 1] = objEnvLayers[i]->get(intPos);

    // put mask value as the last value
    cell->values[intCount + 1] = bytMaskValue;

    // alocate new object
    cell->intCount = intCount + 2;

    return true;
  }

  else
    // point is masked
    return false;
}

// ================================================================================
void EnvLayerSet::getPositionOfPresencePoints()
{
  // read first env layer and create array of presence point positions
  BYTE val;
  int i, nPts, currPos;

  nPts = currPos = 0;

  for (i = 0; i < intCells; i++)
  {
    // get value from file
    val = objSpeciesLayer->get(i);

    // check if it has present value
    if (val == PRESENCE)
      // it is a presence value: add to array
      intPresencePointPosition[nPts++] = currPos;

    // go to next position
    currPos++;
  }

  objSpeciesLayer->rewind();

  intPresencePointCount = nPts;
  blnPresencePointCalculated = true;
}

// ================================================================================
void EnvLayerSet::presampleFromDataPoints(EnvCellSet * cellSet, char * strDataPoints, bool blnEqual, int intResamples, int iBgSurfaceIndex)
{
  // presample data given via string, and count all masked and unmasked points given
  // the other 2 methods (presampleEqual and presampleProportional)

  EnvLayer * speciesLayer;

  // first create a temp envlayer to convert and hold the datapoints
  speciesLayer = new EnvLayer;

  try
  {
    // copy EnvLayerSet metadata into species layer
    metadataToEnvLayer(speciesLayer);

    if (blnEqual)
    {
      // create special type of byte matrix to hold datapoints
      speciesLayer->createMatrix("BooleanByteMatrix", "");
    }
    else
    {
      // no special matrix needed
      // just create an in memory byte matrix
      speciesLayer->createMatrix("RawByteMatrixInMemory", "");
    }

    // set this layer as the species layer
    addAsSpecies(speciesLayer);

    // load speciesLayer with data points
    speciesLayer->fromDataPointStringSettingCounters(strDataPoints, objMaskLayer,
            lMaskedPoints, lUnmaskedPoints, lUniqueMaskedPoints, lUniqueUnmaskedPoints);

    // then call the right presample algorithm
    if (blnEqual)
    {
      // prepare the array that stores the presence position
      ((BooleanByteMatrix *) (speciesLayer->getMatrix()))->copyPresenceArray(intPresencePointCount, intPresencePointPosition);

      blnPresencePointCalculated = true;

      // remove masked points from presence array
      filterMaskedPointsFromPresenceArray();

      if (iBgSurfaceIndex == -1)
      {
        // presample equal
        presampleEqual(cellSet, intResamples);
      }
      else
      {
        // use background surface for resampling bg points
        presampleFromBackgroundSurface(cellSet, intResamples, iBgSurfaceIndex);
      }
    }
    else
    {
      // presample proportional
      presampleProportional(cellSet, intResamples);
    }
  }
  catch (...)
  {
    // remove temp species layer and deallocate it
    delete speciesLayer;
    objSpeciesLayer = NULL;

    throw;
  }

  // remove temp species layer and deallocate it
  delete speciesLayer;
  objSpeciesLayer = NULL;
}

// ================================================================================
void EnvLayerSet::filterMaskedPointsFromPresenceArray()
{
  int i, j;

  for (i = 0; i < intPresencePointCount; i++)
  {
    if (masked(intPresencePointPosition[i]))
    {
      // point is masked
      // remove it by copying everybody to the previous position
      // this will keep the array sorted
      for (j = i; j < intPresencePointCount - 1; j++)
        intPresencePointPosition[j] = intPresencePointPosition[j + 1];

      // there is one element less in the array: decrease count
      intPresencePointCount--;

      // the value that replaced the masked one wasn't tested
      // we don't know whether it is masked or not
      // so decrement i so this element gets verified in the next iteration
      i--;
    }
    else
      i = i;
  }
}

// ================================================================================
void EnvLayerSet::transferDataPointsToCellSet(EnvCellSet * cellSet, char * strDataPoints)
{
  EnvLayer * oTempSpeciesLayer, * oOldSpeciesLayer;
  char * strAux, * strToken;
  double dblLong, dblLat;
  int intPos;


  // check if there is any data point
  if (!strDataPoints)
    throw GarpException(37, "No datapoint found.");

  // create a temp envlayer to convert and hold the datapoints
  oTempSpeciesLayer = new EnvLayer;

  // copy EnvLayerSet metadata into species layer
  metadataToEnvLayer(oTempSpeciesLayer);

  // create special type of byte matrix to hold datapoints
  oTempSpeciesLayer->createMatrix("BooleanByteMatrix", "");

  // load speciesLayer with data points
  oTempSpeciesLayer->fromDataPointString(strDataPoints);

  // set this layer as the species layer
  oOldSpeciesLayer = getSpecies();
  addAsSpecies(oTempSpeciesLayer);

  // copy original datapoints in temp string variable
  strAux = new char[strlen(strDataPoints) + 2];
  strcpy(strAux, strDataPoints);

  try
  {
    strToken = strtok(strAux, " \n\t\r");
    while (strToken)
    {
      // get longitude from token that has already been read
      dblLong = atof(strToken);

      // read latitude from next token
      strToken = strtok(NULL, " \n\t\r");
      if (!strToken)
        throw GarpException(37, "Invalid datapoint file");
      else
        dblLat  = atof(strToken);

      // get position in matrix
      intPos = oTempSpeciesLayer->getPosition(dblLong, dblLat);

      // if point is inside layer extent, put the cell into CellSet
      if (intPos >= 0)
      {
        // get cell from
        EnvCell * cell;
        cell = getCell(intPos);

        // add cell if it is not masked
        if (cell)
          cellSet->add(cell);
      }

      // get next token (longitude of next datapoint)
      strToken = strtok(NULL, " \n\t\r");
    }

    // restore original species layer
    addAsSpecies(oOldSpeciesLayer);

  }

  catch (...)
  {
    throw GarpException(38, "Error reading datapoint string");
  }

  if (strAux)
    delete strAux;

  return;
}

// ================================================================================
void EnvLayerSet::presampleFromBackgroundSurface(EnvCellSet * cellSet, int intResamples, int iBgSurfaceIndex)
{
  int i, j, n, posP, p;
  EnvCell * cell;
  long sum = 0;           // total sum of values on the grid
  long * random = NULL;   // vector with all random numbers generated
  long left;				// how much of the current random value was used during walk
  long currSum = 0;		// cumulative sum when walking through the bg surface
  long temp;              // temp variable used during replacement of values in bubble sort
  long unscaledValue;
  BYTE scaledValue;
  EnvLayer * oBgSurface;

  if (!blnPresencePointCalculated)
    getPositionOfPresencePoints();

  if (!intPresencePointCount)
    throw GarpException(43, "No data points found.");

  n = intResamples / 2;

  // sample presence points (half of sample as in the equal (non-proportional) sample
  for (i = 0; i < n; i++)
  {
    // sample a presence point if there is any
    do
    {
      p = GarpUtil::randint(0, intPresencePointCount);
      posP = intPresencePointPosition[p];

      // continue to get points until a non-masked presence point is found
      // WARNING!!! Does not take into account the replacement of points
    }
    while (masked(posP));

    // get the cell from layers
    cell = getCell(posP);
    cellSet->add(cell);
  }

  // now sample the absences

  // get background surface
  if (iBgSurfaceIndex < intCount)
  {
    oBgSurface = getLayerByIndex(iBgSurfaceIndex);
  }
  else
  {
    // if index overflow, then take the mask
    oBgSurface = getMask();
  }

  // get total sum of all non-masked values on the background surface
  for (i = 0; i < oBgSurface->intCells; i++)
  {
    // mask value counts as zero (i.e., there is no chance of sampling a masked cell)
    if (!masked(i))
    {
      scaledValue = oBgSurface->get(i);
      unscaledValue = (long) oBgSurface->unscale(scaledValue);

      sum += unscaledValue;
    }
  }

  // now get the <n> random values from 0 to <sum> to sample and store in a vector
  random = new long[n];
  for (i = 0; i < n; i++)
  {
    random[i] = GarpUtil::randint(0, sum);
  }

  // sort the <n> values
  // bubble sort implemented as first approach. change to quicksort later
  for (i = 0; i < n - 1; i++)
  {
    for (j = 0; j < n - i - 1; j++)
    {
      if (random[j] > random[j + 1])
      {
        temp = random[j];
        random[j] = random[j + 1];
        random[j + 1] = temp;
      }
    }
  }

  // debug
  int * hits = new int[intCells]; for (i = 0; i < intCells; i++) hits[i] = 0;

  // now walk trhough the background surface getting the cells that the random values select
  // TODO: explain the idea behind this sampling algorithm
  posP = 0;
  currSum = 0;
  for (i = 0; i < n; i++)
  {
    left = random[i] - currSum;

    // while there is still something to subtract from <left> keep moving forward in the vector
    while (left >= 0)
    {
      // mask value counts as zero (i.e., there is no chance of sampling a masked cell)
      if (!masked(posP))
      {
        // get current cell value from bg surface
        scaledValue = oBgSurface->get(posP);
        unscaledValue = (long) oBgSurface->unscale(scaledValue);
      }
      else
      {
        // mask counts as zero
        unscaledValue = 0;
      }

      // subtract the current cell value from <left> and move to the next cell
      left -= unscaledValue;

      // add current value to the current sum
      currSum += unscaledValue;

      // move to the next cell
      posP++;
    }

    // <posP> holds the corresponding position where there was no <left> left

    // get the cell from layers
    cell = getCell(posP - 1);
    cellSet->add(cell);

    // debug
    hits[posP - 1]++;
  }

  //debug
  delete hits;

  // clean up
  delete random;
}

// ================================================================================
void EnvLayerSet::presampleEqual(EnvCellSet * cellSet, int intResamples)
{
  int i, n, posP, p;
  EnvCell * cell;

  if (!blnPresencePointCalculated)
    getPositionOfPresencePoints();

  if (!intPresencePointCount)
    throw GarpException(43, "No data points found.");

  n = intResamples / 2;


  /*// DEBUG: this code extracts data from GARP for use with GRASP
  FILE * ys, * xs, * ps;
  double coordx, coordy, dValue;
  BYTE bValue;
  int j;

  // open files
  ys = fopen("grasper_ys.txt", "w");
  xs = fopen("grasper_xs.txt", "w");
  ps = fopen("grasper_ps.txt", "w");

  // headers
  fprintf(ys, "index\tVOI_name\n");
  fprintf(xs, "index\t long\tlat");
  fprintf(ps, "index\tlong\tlat");
  for (i = 0; i < intCount; i++)
  {
    fprintf(xs, "\t%s", this->objEnvLayers[i]->strTitle);
    fprintf(ps, "\t%s", this->objEnvLayers[i]->strTitle);
  }

  fprintf(xs, "\n");
  fprintf(ps, "\n");

  // DEBUG*/

  // non proportional sampling: sample one absence and one presence at a time
  for (i = 0; i < n; i++)
  {
    // sample an absence point
    do
    {
      // first get a point that is not in the presence array (absence point)
      do
      { posP = GarpUtil::randint(0, intCells); }
      while (GarpUtil::isInArray(&intPresencePointPosition[0], intPresencePointCount, posP));

      // continue to get points until a non-masked absence point is found
    }
    while (masked(posP));

    // get the cell from layers
    cell = getCell(posP);
    cellSet->add(cell);

    /*// DEBUG: this code extracts data from GARP for use with GRASP
    fprintf(ys, "%d\t%f\n", i * 2 + 1, 0.0);
    objMaskLayer->getCoordinates(posP, coordx, coordy);
    fprintf(xs, "%d\t%f\t%f", i * 2 + 1, coordx, coordy);
    for (j = 0; j < intCount; j++)
    {
      bValue = objEnvLayers[j]->get(posP);
      dValue = objEnvLayers[j]->unscale(bValue);
      fprintf(xs, "\t%f", dValue);
    }
    fprintf(xs, "\n");
    // DEBUG*/

    // sample a presence point if there is any
    do
    {
      p = GarpUtil::randint(0, intPresencePointCount);
      posP = intPresencePointPosition[p];

      // continue to get points until a non-masked presence point is found
      // WARNING!!! Does not take into account the replacement of points
    }
    while (masked(posP));

    // get the cell from layers
    cell = getCell(posP);
    cellSet->add(cell);

    /*// DEBUG: this code extracts data from GARP for use with GRASP
    fprintf(ys, "%d\t%f\n", i * 2 + 2, 1.0);
    objMaskLayer->getCoordinates(posP, coordx, coordy);
    fprintf(xs, "%d\t%f\t%f", i * 2 + 2, coordx, coordy);
    for (j = 0; j < intCount; j++)
    {
      bValue = objEnvLayers[j]->get(posP);
      dValue = objEnvLayers[j]->unscale(bValue);
      fprintf(xs, "\t%f", dValue);
    }
    fprintf(xs, "\n");
    // DEBUG*/
  }

  /*// DEBUG: this code extracts data from GARP for use with GRASP
  double dx, dy, dxStep, dyStep;

  dxStep = (dblXMax - dblXMin) / 100;
  dyStep = (dblYMax - dblYMin) / 100;

  int idx = 0;
  for (dx = dblXMin; dx < dblXMax; dx += dxStep)
  {
    for (dy = dblYMin; dy < dblYMax; dy += dyStep)
    {
      posP = objMaskLayer->getPosition(dx, dy);

      if (posP != -1)
      {
        idx++;

        fprintf(ps, "%d\t%f\t%f", idx, dx, dy);

        for (i = 0; i < intCount; i++)
        {
          bValue = objEnvLayers[i]->get(posP);
          dValue = objEnvLayers[i]->unscale(bValue);
          fprintf(ps, "\t%f", dValue);
        }
        fprintf(ps, "\n");
      }
    }
  }

  fclose(ps);
  fclose(ys);
  fclose(xs);
  // DEBUG*/
}

// ================================================================================
void EnvLayerSet::presampleProportional(EnvCellSet * cellSet, int intResamples)
{
  throw GarpException(101, "Method EnvLayerSet::presampleProportional() is deprecated");

  int i, posP;
  EnvCell * cell;

  // sample any non-masked point
  for (i = 0; i < intResamples; i++)
  {
    do
    { posP = GarpUtil::randint(0, intCells); }
    while (masked(posP));

    // get the cell from layers
    cell = getCell(posP);
    cellSet->add(cell);
  }
}

// ================================================================================
void EnvLayerSet::getEnvCombinations(EnvCellSet * cellSet)
{
  int i, j, k, n;
  EnvCell * cell;
  int * oNum = NULL;

  oNum = new int[intCells];
  for (i = 0; i < intCells; i++)
    oNum[i] = 0;

  k = 0;
  for (i = 0; i < intCells; i++)
  {
    if (!masked(i))
    {
      // count the number of non-masked cells
      k++;

      // get the cell from layers
      cell = getCell(i);

      // check if it is equal to any other cell in the set
      j = 0;
      n = cellSet->count();
      while ((j < n) && (*cell != *(cellSet->get(j))))
        j++;

      // if it is a new combination then add it to the cell set
      if (j == n)
        cellSet->add(cell);
      else
        oNum[j]++;
    }
  }

  int sum = 0, max = 0;
  double avg;
  for (i = 0; i < cellSet->count(); i++)
  {
    sum += oNum[i];
    if (oNum[i] > max)
      max = oNum[i];
  }
  avg = (double) sum / cellSet->count();

  if (oNum)
    delete oNum;
}

// ================================================================================
EnvLayer * EnvLayerSet::getLayerByIndex(int intIndex)
{
  if (intIndex > intCount || intIndex < 0)
  {
    throw GarpException(91, "EnvLayerSet::getLayerByIndex(): Index out of range");
	return NULL;
  }
  else
    return objEnvLayers[intIndex];
}

// ================================================================================
EnvLayer * EnvLayerSet::getLayerById(char * strId)
{
  int i;

  for (i = 0; i < intCount; i++)
    if (strcmp(objEnvLayers[i]->strId, strId) == 0)
      return objEnvLayers[i];

  // id not found
  return NULL;
}

// ================================================================================
EnvLayer * EnvLayerSet::getMask()
{ return objMaskLayer; }

// ================================================================================
EnvLayer * EnvLayerSet::getSpecies()
{ return objSpeciesLayer; }

// ================================================================================
// ================================================================================
// ================================================================================

