/**
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.srb;

import ptolemy.actor.NoRoomException;
import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.data.ArrayToken;
import ptolemy.data.ObjectToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.type.ArrayType;
import ptolemy.data.type.BaseType;
import ptolemy.gui.GraphicalMessageHandler;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;

import java.util.Vector;

import edu.sdsc.grid.io.GeneralFile;
import edu.sdsc.grid.io.MetaDataCondition;
import edu.sdsc.grid.io.MetaDataRecordList;
import edu.sdsc.grid.io.MetaDataSelect;
import edu.sdsc.grid.io.MetaDataSet;
import edu.sdsc.grid.io.MetaDataTable;
import edu.sdsc.grid.io.local.LocalFile;
import edu.sdsc.grid.io.srb.SRBFile;
import edu.sdsc.grid.io.srb.SRBFileSystem;
import edu.sdsc.grid.io.srb.SRBMetaDataRecordList;
import edu.sdsc.grid.io.srb.SRBMetaDataSet;

//////////////////////////////////////////////////////////////////////////
//// SRBAddMD
/** Add user defined metadata. When ingesting metadata, only equal operator is used.

   @author Efrat Jaeger
   @version $Id: SRBGetMD.java,v 1.1 2004/10/13 01:34:14 jaeger Exp $
   @category.name srb
   @category.name put

*/
public class SRBGetMD extends TypedAtomicActor {

	/** Construct a constant source with the given container and name.
	 *  Create the <i>value</i> parameter, initialize its value to
	 *  the default value of an IntToken with value 1.
	 *  @param container The container.
	 *  @param name The name of this actor.
	 *  @exception IllegalActionException If the entity cannot be contained
	 *   by the proposed container.
	 *  @exception NameDuplicationException If the container already has an
	 *   actor with this name.
	 */
	public SRBGetMD(CompositeEntity container, String name)
		throws NameDuplicationException, IllegalActionException {
		super(container, name);

	    SRBConnection = new TypedIOPort(this, "SRBFileSystem", true, false);
	    SRBConnection.setTypeEquals(BaseType.GENERAL);
	    new Attribute(SRBConnection, "_showName");

	    srbFilePath = new TypedIOPort(this, "srbFilePath", true, false);
	    srbFilePath.setTypeEquals(BaseType.STRING); // or should it be an array of strings.
	    new Attribute(srbFilePath, "_showName");

	    output = new TypedIOPort(this, "output", false, true);
	    output.setTypeEquals(BaseType.STRING);
	    new Attribute(output, "_showName");

	    exitCode = new TypedIOPort(this, "exitCode", false, true);
	    exitCode.setTypeEquals(BaseType.STRING);
	    new Attribute(exitCode, "_showName");
	    
	    _attachText("_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"60\" height=\"30\" "
                + "style=\"fill:white\"/>\n"
                + "<text x=\"4\" y=\"20\""
                + "style=\"font-size:16; fill:blue; font-family:SansSerif\">"
                + "[SRB]</text>\n"
                + "<text x=\"45\" y=\"22\""
                + "style=\"font-size:20; fill:blue; font-family:SansSerif\">"
                + "$</text>\n"
                + "</svg>\n");
	}

	public TypedIOPort SRBConnection;
	public TypedIOPort srbFilePath;
	public TypedIOPort output;
	public TypedIOPort exitCode;
	
	///////////////////////////////////////////////////////////////////
	////                         public methods                    ////
        /** Get the physical location of SRB logical file paths.
         *  @exception IllegalActionException If it is thrown if the SRB 
         *  file cannot be accessed or the current directory cannot be 
         *  broadcasted.
         */
    public void fire() throws IllegalActionException {
        	
        SRBFile srbFile;
        LocalFile localFile;
        String localFilePath;
        String _exitCode = "";
 
        try {
        	// Get the SRB connection token.
        	SRBFileSystem srbFileSystem = (SRBFileSystem) ( (ObjectToken)
        			SRBConnection.get(0)).getValue();
        	
        	String srbFileStr = ((StringToken)srbFilePath.get(0)).stringValue();
        	srbFile = new SRBFile(srbFileSystem, srbFileStr);
        	
        	if (srbFile.exists()) {
	
	    		String[] selectFieldNames = new String[2];
	    		selectFieldNames[0] = SRBMetaDataSet.FILE_NAME; 
	    		if (srbFile.isDirectory()) {
			    		selectFieldNames[1] = SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES;
			    		//FIXME: How to output metadata non-recursively.
	    		}
	    		else {
		    		selectFieldNames[1] = SRBMetaDataSet.DEFINABLE_METADATA_FOR_FILES;
	    		}

	    		MetaDataSelect selects[] =
	    			MetaDataSet.newSelection( selectFieldNames );

	    		MetaDataRecordList[] record = ((GeneralFile) srbFile).query( selects );
	    		_sendResults( record );
	    		
        	} else GraphicalMessageHandler.error(srbFile.getAbsolutePath() +"does not exist.");
        	
        	//FIXME: what should be returned here..
        	if (_exitCode.equals("")) {
        		_exitCode = "success";
        	}
        	exitCode.broadcast(new StringToken(_exitCode));
        } catch (Exception ex) {
        	ex.printStackTrace();
        }
    }

	/** Post fire the actor. Return false to indicated that the
	 *  process has finished. If it returns true, the process will
	 *  continue indefinitely.
	 */
	public boolean postfire() {
		return true;
	}
	
	/** Outputs the metadata
	 * 
	 * @param rl metadata record list.
	 * @throws IllegalActionException broadcast exception
	 */
	private void _sendResults( MetaDataRecordList[] rl ) throws IllegalActionException
	{
		String results= "";
		if (rl != null) { //Nothing in the database matched the query
			for (int i=0;i<rl.length;i++) {
				results += rl[i] + "\n";
			}
		}	
		output.broadcast(new StringToken(results));
	}

}
