/**
 *    '$RCSfile: EcoGridQueryServicesController.java,v $'
 *
 *     '$Author: tao $'
 *       '$Date: 2004/10/11 23:46:12 $'
 *   '$Revision: 1.3 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.ecogrid;

import java.util.Hashtable;
import java.util.Vector;

import org.apache.xpath.XPathAPI;

import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import org.ecoinformatics.seek.ecogrid.exception.InvalidEcoGridServiceException;
import org.ecoinformatics.util.Config;
import org.ecoinformatics.util.DebugPrinter;

/**
 * This class will controll the current service user want to search:
 * user can add, remove and update the service list
 * @author Jing Tao
 *  
 */

public class EcoGridQueryServicesController extends EcoGridServicesController
{
  private Vector currentQueryServicesList = new Vector();
  private static DebugPrinter dbg         = DebugPrinter.getInstance();

  /**
   * Default constructor. It will read services from configure file first.
   */
  public EcoGridQueryServicesController() 
  {
    readQueryServiceFromConfig();
  }//EcoGridServicesController

  /*
   * Read query service from configure file and put the service into a vector
   */
  private void readQueryServiceFromConfig()
  {
    String xpath  = "//" + MetadataSpecificationInterface.ECOGRIDPATH + "/" + 
                   MetadataSpecificationInterface.SERVICESLIST + "/"+
                   MetadataSpecificationInterface.SERVICE + "/" + 
                   MetadataSpecificationInterface.SERVICETYPE;
    NodeList typeList = Config.getNodeListFromPath(xpath);
    // find service type is ecogrid query 
    if (typeList != null)
    {
      int length = typeList.getLength();
      for (int i=0; i<length; i++)
      {
        Node serviceTypeNode = typeList.item(i);
        if (serviceTypeNode != null && serviceTypeNode.getFirstChild()!= null)
        {
          String serviceTypeString = serviceTypeNode.getFirstChild().getNodeValue();
          dbg.print("The service type from configure file is " + 
                                        serviceTypeString, 2);
          //if this is query service, transfer the service node(servicetype' parent
          // node) into java object and put it into query service list
          if (serviceTypeString != null && 
               serviceTypeString.equals(MetadataSpecificationInterface.QUERYSERVIETYPE))
          {
            Node queryServiceNode = serviceTypeNode.getParentNode();
            EcoGridService queryService = generateServiceFromServiceNode(queryServiceNode);
            if (queryService != null)
            {
              // check duplicate
              boolean duplicate = true;
              try
              {
                duplicate = isDuplicateService(queryService, currentQueryServicesList);
              }
              catch(InvalidEcoGridServiceException e)
              {
                dbg.print("The error for checking duplicate ecogrid service in " +
                          "readQueryServiceFromConfig is " + e.getMessage(), 2);
              }
              if (!duplicate)
              {
                dbg.print("read service " + queryService.getServiceName() +
                          " from config file to vector", 2);
                currentQueryServicesList.add(queryService);
              }
            }
          }//if
        }//if
      }//for
    }//if

  }//readQueryServiceFromConfig

  
 
  /**
   * Remove a given service from current query service list. The key for removing
   * is endpoint. If endpoint is same, we think they are same service
   * @param service EcoGridService
   */
  public void removeService(EcoGridService service)
  {
    if (service == null)
    {
      //if service is null, we don't need do anything
     return;
    }//if
    else
    {
       String endPoint = service.getEndPoint();
       if (endPoint == null || endPoint.trim().equals(""))
       {
         return;
       }//if
       else
       {
          // go through the current query service list and find a servcie
          // which has same endpoint to the given servcie. The remove it.
          int size = currentQueryServicesList.size();
          for (int i=0; i <size; i++)
          {
            EcoGridService currentService = (EcoGridService)
                                          currentQueryServicesList.elementAt(i);
            if (currentService != null)
            {
              String currentEndPoint = currentService.getEndPoint();
              if (currentEndPoint != null && currentEndPoint.equals(endPoint))
              {
                currentQueryServicesList.remove(i);
                dbg.print("Delete service " + service.getServiceName() + 
                          " from list", 2);
                return;
              }//if
            }//if
          }//for
      }//else
    }//else    
  }//removeQueryService

  /**
   * Add a service to query service list. If the service is duplicate or
   * the service is not a query service. It will throw an exception
   * @param service EcoGridService
   * @throws InvalidEcoGridServiceException
   */
  public void addService(EcoGridService service) throws InvalidEcoGridServiceException
  {
     // isDuplicateService will check out if the service is null
     if (!isDuplicateService(service, currentQueryServicesList))
     {
       // make sure the service is query type
       String serviceType = service.getServiceType();
       if (serviceType == null ||
            !serviceType.equals(MetadataSpecificationInterface.QUERYSERVIETYPE))
       {
          throw new InvalidEcoGridServiceException("The service type is not " +
                     "query type and couldn't be add into list");
       }
       dbg.print("Add service "+ service.getServiceName() +" into list", 2);
       currentQueryServicesList.add(service);  
     }
     else
     {
       throw new InvalidEcoGridServiceException("The service has end point " +
                                               service.getEndPoint() + 
         " is already existed in the list and you can't add it into list again"); 
     }
  }//addQueryService
  
  /**
   * Method to update a existed EcoGridService
   * @param service EcoGridService
   * @throws InvalidEcoGridServiceException
   */
  public void updateService(EcoGridService service) throws InvalidEcoGridServiceException
  {
    if (service == null)
    {
      throw new InvalidEcoGridServiceException("Couldn't use a null service" +
                                               " to update a list");
    }
    // make sure the service is query type
    String serviceType = service.getServiceType();
    String endPoint    = service.getEndPoint();
    if (serviceType == null || 
            !serviceType.equals(MetadataSpecificationInterface.QUERYSERVIETYPE))
    {
      throw new InvalidEcoGridServiceException("The service type is not " +
                     "query type and couldn't be updated into list");
    }//if
    if (endPoint == null || endPoint.trim().equals(""))
    {
      throw new InvalidEcoGridServiceException("End point couldn't be null in "+ 
                                  "the new service which will update the list");
    }
    boolean success = false;
    //go through the vector
    int size = currentQueryServicesList.size();
    for (int i=0; i<size; i++)
    {
      EcoGridService currentService = (EcoGridService)currentQueryServicesList.elementAt(i);
      String currentEndPoint = currentService.getEndPoint();
      if (currentEndPoint != null && currentEndPoint.equals(endPoint))
      {
        //remove the old one
        currentQueryServicesList.remove(i);
        currentQueryServicesList.add(service);
        dbg.print("Update the service " + service.getServiceName(), 2);
        success = true;
      }
    }//for
    // couldn't find a service to update and will throw a exception
    if (!success)
    {
      throw new InvalidEcoGridServiceException("Couldn't find a target service " +
                                                     "to update in the list");
    }
  }//updateService
  
  /**
   * Print out every service in the list
   */
  public void print()
  {
    if (currentQueryServicesList != null)
    {
      int size = currentQueryServicesList.size();
      for (int i=0; i<size; i++)
      {
        EcoGridService service = (EcoGridService)currentQueryServicesList.elementAt(i);
        service.print();
      }//for
    }//if
  }//print
  
  /**
   * Method to get current query service list from controller
   * @return Vector
   */
  public Vector getServicesList()
  {
    return this.currentQueryServicesList;
  }
}//EcoGridServiceController
