/**
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.srb;

import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Const;
import ptolemy.actor.lib.Source;
import ptolemy.actor.parameters.PortParameter;
import ptolemy.data.ArrayToken;
import ptolemy.data.BooleanToken;
import ptolemy.data.ObjectToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.UnsignedByteToken;
import ptolemy.data.expr.FileParameter;
import ptolemy.data.type.ArrayType;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;

import java.util.Date;
import java.util.Vector;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.SimpleDateFormat;

import org.geon.PolygonUtil;

import edu.sdsc.grid.io.RemoteFile;
import edu.sdsc.grid.io.local.LocalFile;
import edu.sdsc.grid.io.srb.SRBFile;
import edu.sdsc.grid.io.srb.SRBFileSystem;
import edu.sdsc.grid.io.srb.SRBRandomAccessFile;

//////////////////////////////////////////////////////////////////////////
//// SPut
/**
   Uploads a local file to the SRB.

   @author Efrat Jaeger
   @version $Id: SPut.java,v 1.5 2004/10/08 01:00:15 jaeger Exp $
   @category.name srb
   @category.name put

*/

public class SPut extends TypedAtomicActor {

	/** Construct a constant source with the given container and name.
	 *  Create the <i>value</i> parameter, initialize its value to
	 *  the default value of an IntToken with value 1.
	 *  @param container The container.
	 *  @param name The name of this actor.
	 *  @exception IllegalActionException If the entity cannot be contained
	 *   by the proposed container.
	 *  @exception NameDuplicationException If the container already has an
	 *   actor with this name.
	 */
	public SPut(CompositeEntity container, String name)
		throws NameDuplicationException, IllegalActionException {
		super(container, name);

	    SRBConnection = new TypedIOPort(this, "SRBFileSystem", true, false);
	    SRBConnection.setTypeEquals(BaseType.GENERAL);
	    new Attribute(SRBConnection, "_showName");

	    filesToPut = new TypedIOPort(this, "filesToPut", true, false);
	    filesToPut.setTypeEquals(new ArrayType(BaseType.STRING));
	    new Attribute(filesToPut, "_showName");

	    uploadedFiles = new TypedIOPort(this, "uploadedFiles", false, true);
	    uploadedFiles.setTypeEquals(new ArrayType(BaseType.STRING));
	    new Attribute(uploadedFiles, "_showName");
	    
	    exitCode = new TypedIOPort(this, "exitCode", false, true);
	    exitCode.setTypeEquals(BaseType.STRING);
	    new Attribute(exitCode, "_showName");
	    
	    remoteDir = new PortParameter(this, "remoteDir");
	    new Attribute(remoteDir, "_showName");

	    _attachText("_iconDescription", "<svg>\n"
	            + "<rect x=\"-25\" y=\"-20\" "
	            + "width=\"50\" height=\"40\" "
	            + "style=\"fill:white\"/>\n"
	            + "<polygon points=\"-15,-10 -12,-10 -8,-14 -1,-14 3,-10"
	            + " 15,-10 15,10, -15,10\" "
	            + "style=\"fill:red\"/>\n"
	            + "<text x=\"-13.5\" y=\"7\" "
	            + "style=\"font-size:14\">\n"
	            + "SRB \n"
	            + "</text>\n"
	            + "</svg>\n");
	}

	public TypedIOPort SRBConnection;
	public TypedIOPort uploadedFiles;
	public TypedIOPort filesToPut;
	public TypedIOPort exitCode;
	public PortParameter remoteDir;

	///////////////////////////////////////////////////////////////////
	////                         public methods                    ////
        /** Upload the file to the SRB. If the SRB file path is not
         *  specified, upload to the current working directory. Output 
         *  the current working directory.
         *  @exception IllegalActionException If it is thrown if the SRB 
         *  file cannot be accessed or the current directory cannot be 
         *  broadcasted.
         */
    public void fire() throws IllegalActionException {
        	
        SRBFile srbFile;
        LocalFile localFile;
        String srbFilePath;
        String _exitCode = "";
 
        try {
        	// Get the SRB connection token.
        	SRBFileSystem srbFileSystem = (SRBFileSystem) ( (ObjectToken)
        			SRBConnection.get(0)).getValue();
        	
        	// The srb remote directory to fetch the files.
        	remoteDir.update();
        	String remoteDirStr = ((StringToken)remoteDir.getToken()).stringValue();
        	SRBFile parent = new SRBFile(srbFileSystem, remoteDirStr);

        	// Getting the local files list token and copying each file to SRB.
        	ArrayToken localFilesTokenArr = null;
        	try {
        		localFilesTokenArr = (ArrayToken) filesToPut.get(0);
        	} catch (Exception ex) {
        		_debug("filesToGet port is null.");
            }
        	if (localFilesTokenArr != null) {
        		Token[] localFilesToken = localFilesTokenArr.arrayValue();
        		Vector uploadedFilesVec = new Vector();
        		for (int i=0; i<localFilesToken.length; i++) {

        			// local file path.
        			String localFileStr = ((StringToken) localFilesToken[i]).stringValue();
        			_debug("<FILE_TO_PUT>" + localFileStr + "<FILE_TO_PUT>");

                	localFile = new LocalFile(localFileStr);
//        			if (localFile.exists()) {
	                	// setting the remote file path.
	        			int slashIndex = localFileStr.lastIndexOf('/');
	        			String tmpFileNameString = localFileStr;
	        			if (slashIndex > -1) {
	        		    	tmpFileNameString = localFileStr.substring(slashIndex + 1);
	        			}
	        		    _debug("<FILE_NAME_STR>" + tmpFileNameString + "</FILE_NAME_STR>");
	        		    
	        		    // copying the local file to SRB.
	        		    srbFile = new SRBFile(parent, tmpFileNameString);
	        		    try {
	        		    	srbFile.copyFrom( localFile );
	        		    	uploadedFilesVec.add(new StringToken(srbFile.getAbsolutePath()));
	        		    } catch(Exception ex) {
	        		    	// If the paralel put fails try to do a stream put.
	        		    	System.out.println("Paralel put failed due to " + ex.getMessage());
	        		    	System.out.println("Trying stream put.");
	        		    	try {
	        		    		_streamPut(srbFile, localFile.getAbsolutePath());
	        		    		//adding the successfully uploaded files paths to the uploaded files array.
		        		    	uploadedFilesVec.add(new StringToken(srbFile.getAbsolutePath()));
	        		    	} catch (Exception stex) {
	        		    		stex.printStackTrace();
	        		    		// even if there is an exception output the successfully uploaded files.
	        		    		System.out.println("failed to upload file " + localFile.getAbsolutePath()
	        		    				+ " to " + parent.getAbsolutePath() + ".\n");
	        		    		_exitCode += "unable to upload file " + localFile.getAbsolutePath();
	        		    		_exitCode += " to " + parent.getAbsolutePath() + ".\n";
	        		    	}
	        		    }
	       // 		} else {
	        //			System.out.println("file " + localFile.getAbsolutePath() + " does not exist.");
	        	//		_exitCode += "file " + localFile.getAbsolutePath() + " does not exist.\n";
	        		//}
        		}
        		Token[] uploadedFilesArr = new Token[uploadedFilesVec.size()];
        		uploadedFilesVec.toArray(uploadedFilesArr);
        		//broadcast the array only if it's non-empty.
        		if (uploadedFilesArr.length > 0) { 
        			uploadedFiles.broadcast(new ArrayToken(uploadedFilesArr));
        		} else {
        			_exitCode = "no files were uploaded.";
        		}
        		if (_exitCode.equals("")) {
        			_exitCode = "success";
        		}
        		exitCode.broadcast(new StringToken(_exitCode));
        	} else {
        		throw new IllegalActionException("No files to put.");
        	}
        } catch (Exception ex) {
        	srbFile = null;
        	throw new IllegalActionException(this, ex.getMessage());
        }
    
    }

	/** Post fire the actor. Return false to indicated that the
	 *  process has finished. If it returns true, the process will
	 *  continue indefinitely.
	 */
	public boolean postfire() {
		return true;
	}
	
	private void _streamPut(SRBFile srbFile, String localFilePath) throws Exception {

		SRBRandomAccessFile srbRandomAccessFile= null;
		byte[] bytesRead = new byte[20000];;
		int nBytesRead;
		InputStream in = new FileInputStream(localFilePath);
		srbRandomAccessFile = new SRBRandomAccessFile( srbFile, "rw");
		
        nBytesRead = in.read(bytesRead);
		while (nBytesRead > 0) {
			srbRandomAccessFile.write(bytesRead);
	        nBytesRead = in.read(bytesRead);
		}			
		in.close();
	}

}
