// query metadata of a file of a collection, query by constraints.

/**
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.srb;

import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.data.ArrayToken;
import ptolemy.data.ObjectToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.type.ArrayType;
import ptolemy.data.type.BaseType;
import ptolemy.gui.GraphicalMessageHandler;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;

import java.util.StringTokenizer;
import java.util.Vector;

import edu.sdsc.grid.io.GeneralFile;
import edu.sdsc.grid.io.MetaDataCondition;
import edu.sdsc.grid.io.MetaDataRecordList;
import edu.sdsc.grid.io.MetaDataSelect;
import edu.sdsc.grid.io.MetaDataSet;
import edu.sdsc.grid.io.MetaDataTable;
import edu.sdsc.grid.io.local.LocalFile;
import edu.sdsc.grid.io.srb.SRBFile;
import edu.sdsc.grid.io.srb.SRBFileSystem;
import edu.sdsc.grid.io.srb.SRBMetaDataRecordList;
import edu.sdsc.grid.io.srb.SRBMetaDataSet;

//////////////////////////////////////////////////////////////////////////
//// SRBAddMD
/** Add user defined metadata. When ingesting metadata, only equal operator is used.

   @author Efrat Jaeger
   @version $Id: SRBQueryMD.java,v 1.3 2004/10/15 05:30:37 jaeger Exp $
   @category.name srb
   @category.name put

*/
public class SRBQueryMD extends TypedAtomicActor {

	/** Construct a constant source with the given container and name.
	 *  Create the <i>value</i> parameter, initialize its value to
	 *  the default value of an IntToken with value 1.
	 *  @param container The container.
	 *  @param name The name of this actor.
	 *  @exception IllegalActionException If the entity cannot be contained
	 *   by the proposed container.
	 *  @exception NameDuplicationException If the container already has an
	 *   actor with this name.
	 */
	public SRBQueryMD(CompositeEntity container, String name)
		throws NameDuplicationException, IllegalActionException {
		super(container, name);

	    SRBConnection = new TypedIOPort(this, "SRBFileSystem", true, false);
	    SRBConnection.setTypeEquals(BaseType.GENERAL);
	    new Attribute(SRBConnection, "_showName");

	    srbFilePath = new TypedIOPort(this, "srbFilePath", true, false);
	    srbFilePath.setTypeEquals(BaseType.STRING); // or should it be an array of strings.
	    new Attribute(srbFilePath, "_showName");

	    conditions = new TypedIOPort(this, "conditions", true, false);
	    conditions.setTypeEquals(new ArrayType(BaseType.STRING));
	    new Attribute(conditions, "_showName");

	    exitCode = new TypedIOPort(this, "exitCode", false, true);
	    exitCode.setTypeEquals(BaseType.STRING);
	    new Attribute(exitCode, "_showName");
	    
	    _attachText("_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"60\" height=\"30\" "
                + "style=\"fill:white\"/>\n"
                + "<text x=\"4\" y=\"20\""
                + "style=\"font-size:16; fill:blue; font-family:SansSerif\">"
                + "[SRB]</text>\n"
                + "<text x=\"45\" y=\"22\""
                + "style=\"font-size:20; fill:blue; font-family:SansSerif\">"
                + "$</text>\n"
                + "</svg>\n");
	}

	public TypedIOPort SRBConnection;
	public TypedIOPort srbFilePath;
	public TypedIOPort conditions;
	public TypedIOPort exitCode;
	
	///////////////////////////////////////////////////////////////////
	////                         public methods                    ////
        /** Get the physical location of SRB logical file paths.
         *  @exception IllegalActionException If it is thrown if the SRB 
         *  file cannot be accessed or the current directory cannot be 
         *  broadcasted.
         */
    public void fire() throws IllegalActionException {
        	
        SRBFile srbFile;
        LocalFile localFile;
        String localFilePath;
        String _exitCode = "";
 
        try {
        	// Get the SRB connection token.
        	SRBFileSystem srbFileSystem = (SRBFileSystem) ( (ObjectToken)
        			SRBConnection.get(0)).getValue();
        	
        	String srbFileStr = ((StringToken)srbFilePath.get(0)).stringValue();
        	srbFile = new SRBFile(srbFileSystem, srbFileStr);
        	
        	if (srbFile.exists()) {
	        	Token[] conds = ((ArrayToken)conditions.get(0)).arrayValue();
	        	int numConds = conds.length;
	    		//String[][] definableMetaDataValues = new String[numConds][2];
	    		//int operators[] = new int[numConds];
	    		
	    		MetaDataCondition conditions[] = new MetaDataCondition[conds.length-1];

	    		boolean userDefined = true;
	    		for (int i=0; i<numConds; i++) {
	    			String condition = ((StringToken)conds[i]).stringValue();
	    			if (condition.equals("|")) {
	    				userDefined = false;
	    				continue; //separator between user define and dataset conditions.
	    			}
	    			
	    			StringTokenizer st = new StringTokenizer(condition);
	    			int j = 0;
	    			
	    			String att = st.nextToken();
	    			if (!userDefined) 
	    				att = _getAttribute(att);
	    			if (att.equals(""))
	    				GraphicalMessageHandler.error("query attribute is null");
	    			String opStr = st.nextToken();
	    			int op = _getOperator(opStr);
	    			if (op <= MetaDataCondition.GREATER_OR_EQUAL) { 
	    				// this is an equality term.
	    				String val = st.nextToken();
	    				conditions[i] = MetaDataSet.newCondition(att, op, val);

	    				if (st.hasMoreTokens()) 
	    					GraphicalMessageHandler.error("Invalid condition '" + condition +"'.");
	    			} else if (op >= MetaDataCondition.IN && 
	    					op <= MetaDataCondition.NOT_IN) {
	    				String val = st.nextToken();
	    				if (opStr.equals("in") || opStr.equals("not_in")) {
	    					// create an array of val
	    				} else { // this is contains.
	    				    // create an array of val??
	    					// create an array of att.
	    				}
	    			} else if (op >= MetaDataCondition.BETWEEN && 
	    					op <= MetaDataCondition.NOT_BETWEEN) {
	    				String val1 = st.nextToken();
	    				String val2 = st.nextToken();
	    				conditions[i] = MetaDataSet.newCondition(att, op, val1, val2);
	    			} else if ((op >= MetaDataCondition.LIKE && 
	    					op <= MetaDataCondition.NOT_LIKE) ||
	    					(op >= MetaDataCondition.SOUNDS_LIKE &&
	    							op <= MetaDataCondition.SOUNDS_NOT_LIKE)) {
	    				String val = st.nextToken();
	    				conditions[i] = MetaDataSet.newCondition(att, op, val);
	    			}
	    		}
	    		// query for the conditions and the select fields.
	
				String[] selectFieldNames = {
						SRBMetaDataSet.DEFINABLE_METADATA_FOR_DIRECTORIES,
						SRBMetaDataSet.DEFINABLE_METADATA_FOR_FILES
				};
	    		MetaDataSelect selects[] =
	    			MetaDataSet.newSelection( selectFieldNames );

	    		MetaDataRecordList[] rl = ((GeneralFile) srbFile).query( conditions, selects );
	    		printQueryResults( rl );

        	} else GraphicalMessageHandler.error(srbFile.getAbsolutePath() +"does not exist.");
        	
        	//FIXME: what should be returned here..
        	if (_exitCode.equals("")) {
        		_exitCode = "success";
        	}
        	exitCode.broadcast(new StringToken(_exitCode));
        } catch (Exception ex) {
        	ex.printStackTrace();
        }
    }

	/** Post fire the actor. Return false to indicated that the
	 *  process has finished. If it returns true, the process will
	 *  continue indefinitely.
	 */
	public boolean postfire() {
		return true;
	}
	
	private String _getAttribute(String att) {
		att = att.trim().toLowerCase();
        if (att.equals("annotation"))
        	return SRBMetaDataSet.FILE_ANNOTATION;
        if (att.equals("annotator"))
        	return SRBMetaDataSet.FILE_ANNOTATION_USERNAME;
        if (att.equals("owner"))
        	return SRBMetaDataSet.OWNER;
        if (att.equals("dataName"))
        	return SRBMetaDataSet.FILE_NAME;
        if (att.equals("collName"))
        	return SRBMetaDataSet.DIRECTORY_NAME;
        if (att.equals("dataType"))
        	return SRBMetaDataSet.FILE_TYPE_NAME;
		return "";
	}
	
	
	private int _getOperator(String op) {
		// currently only equality operators are supported.
		// if no operator is found assume it is equal.
		op = op.trim().toLowerCase();
		
		if (op.equals("="))
			return MetaDataCondition.EQUAL;
		if (op.equals("!=") || op.equals("not="))
			return MetaDataCondition.NOT_EQUAL;
		if (op.equals("<") || op.equals("lt") || op.equals("num_lt"))
			return MetaDataCondition.LESS_THAN;
		if (op.equals("<=") || op.equals("le") || op.equals("num_le"))
			return MetaDataCondition.LESS_OR_EQUAL;
		if (op.equals(">") || op.equals("gt") || op.equals("num_gt"))
			return MetaDataCondition.GREATER_THAN;
		if (op.equals(">=") || op.equals("ge") || op.equals("num_ge"))
			return MetaDataCondition.GREATER_OR_EQUAL;
		if (op.equals("in"))
			return MetaDataCondition.IN;
		if (op.equals("notin") || op.equals("not_in"))
			return MetaDataCondition.NOT_IN;
		if (op.equals("between") || op.equals("num_between"))
			return MetaDataCondition.BETWEEN;
		if (op.equals("notbetween") || op.equals("not_between") || op.equals("num_not_between"))
			return MetaDataCondition.NOT_BETWEEN;
		if (op.equals("like"))
			return MetaDataCondition.LIKE;
		if (op.equals("notlike") || op.equals("not_like"))
			return MetaDataCondition.NOT_LIKE;
		if (op.equals("in") || op.equals("contains"))
			return MetaDataCondition.IN;
		if (op.equals("notin") || op.equals("not_in") || op.equals("not_contains"))
			return MetaDataCondition.NOT_IN;
		if (op.equals("sounds_like"))
			return MetaDataCondition.SOUNDS_LIKE;
		if (op.equals("sounds_not_like") || op.equals("sounds_not_like"))
			return MetaDataCondition.SOUNDS_NOT_LIKE;
		System.out.println("operator " + op + " is unsupported. Treated as equal.");
		return MetaDataCondition.EQUAL;
	}
	
	private void printQueryResults( MetaDataRecordList[] rl )
	{
		//print query results
		if (rl != null) { //Nothing in the database matched the query
				System.out.print("\n");
				for (int i=0;i<rl.length;i++) {
					System.out.print("\n"+rl[i]);
				}
		}
		System.out.println("\n");

	}


}
